@@ -18,10 +18,13 @@
 
 #include <grpc/support/port_platform.h>
 
+#include "src/core/ext/filters/load_reporting/server_load_reporting_filter.h"
+
 #include <string.h>
 
 #include <string>
 
+#include "absl/strings/ascii.h"
 #include "absl/strings/str_format.h"
 
 #include <grpc/grpc_security.h>
@@ -31,10 +34,9 @@
 
 #include "src/core/ext/filters/client_channel/lb_policy/grpclb/grpclb.h"
 #include "src/core/ext/filters/load_reporting/registered_opencensus_objects.h"
-#include "src/core/ext/filters/load_reporting/server_load_reporting_filter.h"
+#include "src/core/lib/address_utils/parse_address.h"
 #include "src/core/lib/channel/channel_args.h"
 #include "src/core/lib/channel/context.h"
-#include "src/core/lib/iomgr/parse_address.h"
 #include "src/core/lib/iomgr/resolve_address.h"
 #include "src/core/lib/iomgr/sockaddr.h"
 #include "src/core/lib/iomgr/socket_utils.h"
@@ -50,7 +52,7 @@ constexpr char kEncodedIpv6AddressLengthString[] = "32";
 constexpr char kEmptyAddressLengthString[] = "00";
 constexpr size_t kLengthPrefixSize = 2;
 
-grpc_error* ServerLoadReportingChannelData::Init(
+grpc_error_handle ServerLoadReportingChannelData::Init(
     grpc_channel_element* /* elem */, grpc_channel_element_args* args) {
   GPR_ASSERT(!args->is_last);
   // Find and record the peer_identity.
@@ -72,7 +74,7 @@ grpc_error* ServerLoadReportingChannelData::Init(
 
 void ServerLoadReportingCallData::Destroy(
     grpc_call_element* elem, const grpc_call_final_info* final_info,
-    grpc_closure* then_call_closure) {
+    grpc_closure* /*then_call_closure*/) {
   ServerLoadReportingChannelData* chand =
       reinterpret_cast<ServerLoadReportingChannelData*>(elem->channel_data);
   // Only record an end if we've recorded its corresponding start, which is
@@ -92,14 +94,13 @@ void ServerLoadReportingCallData::Destroy(
         {{::grpc::load_reporter::TagKeyToken(),
           {client_ip_and_lr_token_, client_ip_and_lr_token_len_}},
          {::grpc::load_reporter::TagKeyHost(),
-          {target_host_, target_host_len_}},
+          {target_host_.data(), target_host_.length()}},
          {::grpc::load_reporter::TagKeyUserId(),
           {chand->peer_identity(), chand->peer_identity_len()}},
          {::grpc::load_reporter::TagKeyStatus(),
           GetStatusTagForStatus(final_info->final_status)}});
     gpr_free(client_ip_and_lr_token_);
   }
-  gpr_free(target_host_);
   grpc_slice_unref_internal(service_method_);
 }
 
@@ -114,12 +115,24 @@ void ServerLoadReportingCallData::StartTransportStreamOpBatch(
     original_recv_initial_metadata_ready_ = op->recv_initial_metadata_ready();
     // Substitute the original closure for the wrapper closure.
     op->set_recv_initial_metadata_ready(&recv_initial_metadata_ready_);
-  } else if (op->send_trailing_metadata() != nullptr) {
-    GRPC_LOG_IF_ERROR(
-        "server_load_reporting_filter",
-        grpc_metadata_batch_filter(op->send_trailing_metadata()->batch(),
-                                   SendTrailingMetadataFilter, elem,
-                                   "send_trailing_metadata filtering error"));
+  }
+  if (op->send_trailing_metadata() != nullptr) {
+    const auto& costs = op->send_trailing_metadata()->batch()->Take(
+        grpc_core::LbCostBinMetadata());
+    for (const auto& cost : costs) {
+      ServerLoadReportingChannelData* chand =
+          reinterpret_cast<ServerLoadReportingChannelData*>(elem->channel_data);
+      opencensus::stats::Record(
+          {{::grpc::load_reporter::MeasureOtherCallMetric(), cost.cost}},
+          {{::grpc::load_reporter::TagKeyToken(),
+            {client_ip_and_lr_token_, client_ip_and_lr_token_len_}},
+           {::grpc::load_reporter::TagKeyHost(),
+            {target_host_.data(), target_host_.length()}},
+           {::grpc::load_reporter::TagKeyUserId(),
+            {chand->peer_identity(), chand->peer_identity_len()}},
+           {::grpc::load_reporter::TagKeyMetricName(),
+            {cost.name.data(), cost.name.length()}}});
+    }
   }
   grpc_call_next_op(elem, op->op());
 }
@@ -134,18 +147,18 @@ std::string ServerLoadReportingCallData::GetCensusSafeClientIpString() {
             "metadata.");
     return "";
   }
-  // Parse the client URI string into grpc_uri.
-  grpc_uri* client_uri = grpc_uri_parse(client_uri_str, true);
-  if (client_uri == nullptr) {
+  absl::StatusOr<grpc_core::URI> client_uri =
+      grpc_core::URI::Parse(client_uri_str);
+  if (!client_uri.ok()) {
     gpr_log(GPR_ERROR,
             "Unable to parse the client URI string (peer string) to a client "
-            "URI.");
+            "URI. Error: %s",
+            client_uri.status().ToString().c_str());
     return "";
   }
   // Parse the client URI into grpc_resolved_address.
   grpc_resolved_address resolved_address;
-  bool success = grpc_parse_uri(client_uri, &resolved_address);
-  grpc_uri_destroy(client_uri);
+  bool success = grpc_parse_uri(*client_uri, &resolved_address);
   if (!success) {
     gpr_log(GPR_ERROR,
             "Unable to parse client URI into a grpc_resolved_address.");
@@ -199,52 +212,38 @@ void ServerLoadReportingCallData::StoreClientIpAndLrToken(const char* lr_token,
   if (lr_token_len != 0) {
     strncpy(cur_pos, lr_token, lr_token_len);
   }
-  GPR_ASSERT(cur_pos + lr_token_len - client_ip_and_lr_token_ ==
-             client_ip_and_lr_token_len_);
+  GPR_ASSERT(
+      static_cast<size_t>(cur_pos + lr_token_len - client_ip_and_lr_token_) ==
+      client_ip_and_lr_token_len_);
 }
 
-grpc_filtered_mdelem ServerLoadReportingCallData::RecvInitialMetadataFilter(
-    void* user_data, grpc_mdelem md) {
-  grpc_call_element* elem = reinterpret_cast<grpc_call_element*>(user_data);
-  ServerLoadReportingCallData* calld =
-      reinterpret_cast<ServerLoadReportingCallData*>(elem->call_data);
-  if (grpc_slice_eq(GRPC_MDKEY(md), GRPC_MDSTR_PATH)) {
-    calld->service_method_ = grpc_slice_ref_internal(GRPC_MDVALUE(md));
-  } else if (calld->target_host_ == nullptr &&
-             grpc_slice_eq(GRPC_MDKEY(md), GRPC_MDSTR_AUTHORITY)) {
-    grpc_slice target_host_slice = GRPC_MDVALUE(md);
-    calld->target_host_len_ = GRPC_SLICE_LENGTH(target_host_slice);
-    calld->target_host_ =
-        reinterpret_cast<char*>(gpr_zalloc(calld->target_host_len_));
-    for (size_t i = 0; i < calld->target_host_len_; ++i) {
-      calld->target_host_[i] = static_cast<char>(
-          tolower(GRPC_SLICE_START_PTR(target_host_slice)[i]));
-    }
-  } else if (grpc_slice_str_cmp(GRPC_MDKEY(md),
-                                grpc_core::kGrpcLbLbTokenMetadataKey) == 0) {
-    if (calld->client_ip_and_lr_token_ == nullptr) {
-      calld->StoreClientIpAndLrToken(
-          reinterpret_cast<const char*> GRPC_SLICE_START_PTR(GRPC_MDVALUE(md)),
-          GRPC_SLICE_LENGTH(GRPC_MDVALUE(md)));
-    }
-    return GRPC_FILTERED_REMOVE();
-  }
-  return GRPC_FILTERED_MDELEM(md);
-}
-
-void ServerLoadReportingCallData::RecvInitialMetadataReady(void* arg,
-                                                           grpc_error* err) {
+void ServerLoadReportingCallData::RecvInitialMetadataReady(
+    void* arg, grpc_error_handle err) {
   grpc_call_element* elem = reinterpret_cast<grpc_call_element*>(arg);
   ServerLoadReportingCallData* calld =
       reinterpret_cast<ServerLoadReportingCallData*>(elem->call_data);
   ServerLoadReportingChannelData* chand =
       reinterpret_cast<ServerLoadReportingChannelData*>(elem->channel_data);
   if (err == GRPC_ERROR_NONE) {
-    GRPC_LOG_IF_ERROR(
-        "server_load_reporting_filter",
-        grpc_metadata_batch_filter(calld->recv_initial_metadata_,
-                                   RecvInitialMetadataFilter, elem,
-                                   "recv_initial_metadata filtering error"));
+    if (const grpc_core::Slice* path =
+            calld->recv_initial_metadata_->get_pointer(
+                grpc_core::HttpPathMetadata())) {
+      calld->service_method_ = path->Ref().TakeCSlice();
+    }
+    if (const grpc_core::Slice* authority =
+            calld->recv_initial_metadata_->get_pointer(
+                grpc_core::HttpAuthorityMetadata())) {
+      calld->target_host_ = absl::AsciiStrToLower(authority->as_string_view());
+    }
+    std::string buffer;
+    auto lb_token =
+        calld->recv_initial_metadata_->Take(grpc_core::LbTokenMetadata());
+    if (lb_token.has_value()) {
+      if (calld->client_ip_and_lr_token_ == nullptr) {
+        calld->StoreClientIpAndLrToken(
+            reinterpret_cast<const char*>(lb_token->data()), lb_token->size());
+      }
+    }
     // If the LB token was not found in the recv_initial_metadata, only the
     // client IP part will be recorded (with an empty LB token).
     if (calld->client_ip_and_lr_token_ == nullptr) {
@@ -255,7 +254,7 @@ void ServerLoadReportingCallData::RecvInitialMetadataReady(void* arg,
         {{::grpc::load_reporter::TagKeyToken(),
           {calld->client_ip_and_lr_token_, calld->client_ip_and_lr_token_len_}},
          {::grpc::load_reporter::TagKeyHost(),
-          {calld->target_host_, calld->target_host_len_}},
+          {calld->target_host_.data(), calld->target_host_.length()}},
          {::grpc::load_reporter::TagKeyUserId(),
           {chand->peer_identity(), chand->peer_identity_len()}}});
   }
@@ -264,65 +263,28 @@ void ServerLoadReportingCallData::RecvInitialMetadataReady(void* arg,
                           GRPC_ERROR_REF(err));
 }
 
-grpc_error* ServerLoadReportingCallData::Init(
-    grpc_call_element* elem, const grpc_call_element_args* args) {
+grpc_error_handle ServerLoadReportingCallData::Init(
+    grpc_call_element* elem, const grpc_call_element_args* /*args*/) {
   service_method_ = grpc_empty_slice();
   GRPC_CLOSURE_INIT(&recv_initial_metadata_ready_, RecvInitialMetadataReady,
                     elem, grpc_schedule_on_exec_ctx);
   return GRPC_ERROR_NONE;
 }
 
-grpc_filtered_mdelem ServerLoadReportingCallData::SendTrailingMetadataFilter(
-    void* user_data, grpc_mdelem md) {
-  grpc_call_element* elem = reinterpret_cast<grpc_call_element*>(user_data);
-  ServerLoadReportingCallData* calld =
-      reinterpret_cast<ServerLoadReportingCallData*>(elem->call_data);
-  ServerLoadReportingChannelData* chand =
-      reinterpret_cast<ServerLoadReportingChannelData*>(elem->channel_data);
-  if (grpc_slice_eq(GRPC_MDKEY(md), GRPC_MDSTR_LB_COST_BIN)) {
-    const grpc_slice value = GRPC_MDVALUE(md);
-    const size_t cost_entry_size = GRPC_SLICE_LENGTH(value);
-    if (cost_entry_size < sizeof(double)) {
-      gpr_log(GPR_ERROR,
-              "Cost metadata value too small (%zu bytes) to hold valid data. "
-              "Ignoring.",
-              cost_entry_size);
-      return GRPC_FILTERED_REMOVE();
-    }
-    const double* cost_entry_ptr =
-        reinterpret_cast<const double*>(GRPC_SLICE_START_PTR(value));
-    double cost_value = *cost_entry_ptr++;
-    const char* cost_name = reinterpret_cast<const char*>(cost_entry_ptr);
-    const size_t cost_name_len = cost_entry_size - sizeof(double);
-    opencensus::stats::Record(
-        {{::grpc::load_reporter::MeasureOtherCallMetric(), cost_value}},
-        {{::grpc::load_reporter::TagKeyToken(),
-          {calld->client_ip_and_lr_token_, calld->client_ip_and_lr_token_len_}},
-         {::grpc::load_reporter::TagKeyHost(),
-          {calld->target_host_, calld->target_host_len_}},
-         {::grpc::load_reporter::TagKeyUserId(),
-          {chand->peer_identity(), chand->peer_identity_len()}},
-         {::grpc::load_reporter::TagKeyMetricName(),
-          {cost_name, cost_name_len}}});
-    return GRPC_FILTERED_REMOVE();
-  }
-  return GRPC_FILTERED_MDELEM(md);
-}
-
 const char* ServerLoadReportingCallData::GetStatusTagForStatus(
     grpc_status_code status) {
   switch (status) {
     case GRPC_STATUS_OK:
-      return ::grpc::load_reporter::kCallStatusOk;
+      return grpc::load_reporter::kCallStatusOk;
     case GRPC_STATUS_UNKNOWN:
     case GRPC_STATUS_DEADLINE_EXCEEDED:
     case GRPC_STATUS_UNIMPLEMENTED:
     case GRPC_STATUS_INTERNAL:
     case GRPC_STATUS_UNAVAILABLE:
     case GRPC_STATUS_DATA_LOSS:
-      return ::grpc::load_reporter::kCallStatusServerError;
+      return grpc::load_reporter::kCallStatusServerError;
     default:
-      return ::grpc::load_reporter::kCallStatusClientError;
+      return grpc::load_reporter::kCallStatusClientError;
   }
 }
 
@@ -339,21 +301,21 @@ bool MaybeAddServerLoadReportingFilter(const grpc_channel_args& args) {
 // time if we build with the filter target.
 struct ServerLoadReportingFilterStaticRegistrar {
   ServerLoadReportingFilterStaticRegistrar() {
-    static grpc_core::Atomic<bool> registered{false};
-    if (registered.Load(grpc_core::MemoryOrder::ACQUIRE)) return;
+    static std::atomic<bool> registered{false};
+    if (registered.load(std::memory_order_acquire)) return;
     RegisterChannelFilter<ServerLoadReportingChannelData,
                           ServerLoadReportingCallData>(
         "server_load_reporting", GRPC_SERVER_CHANNEL, INT_MAX,
         MaybeAddServerLoadReportingFilter);
     // Access measures to ensure they are initialized. Otherwise, we can't
     // create any valid view before the first RPC.
-    ::grpc::load_reporter::MeasureStartCount();
-    ::grpc::load_reporter::MeasureEndCount();
-    ::grpc::load_reporter::MeasureEndBytesSent();
-    ::grpc::load_reporter::MeasureEndBytesReceived();
-    ::grpc::load_reporter::MeasureEndLatencyMs();
-    ::grpc::load_reporter::MeasureOtherCallMetric();
-    registered.Store(true, grpc_core::MemoryOrder::RELEASE);
+    grpc::load_reporter::MeasureStartCount();
+    grpc::load_reporter::MeasureEndCount();
+    grpc::load_reporter::MeasureEndBytesSent();
+    grpc::load_reporter::MeasureEndBytesReceived();
+    grpc::load_reporter::MeasureEndLatencyMs();
+    grpc::load_reporter::MeasureOtherCallMetric();
+    registered.store(true, std::memory_order_release);
   }
 } server_load_reporting_filter_static_registrar;
 
