@@ -15,21 +15,19 @@
  * limitations under the License.
  *
  */
-
 #ifndef GRPC_CORE_LIB_SURFACE_INIT_H
 #define GRPC_CORE_LIB_SURFACE_INIT_H
 
-#ifdef __cplusplus
-extern "C" {
-#endif
+#include <grpc/support/port_platform.h>
 
-void grpc_register_security_filters(void);
-void grpc_security_pre_init(void);
-void grpc_security_init(void);
-int grpc_is_initialized(void);
+#include <memory>
 
-#ifdef __cplusplus
+namespace grpc_event_engine {
+namespace experimental {
+class EventEngine;
 }
-#endif
+}  // namespace grpc_event_engine
+
+void grpc_maybe_wait_for_async_shutdown(void);
 
 #endif /* GRPC_CORE_LIB_SURFACE_INIT_H */
