@@ -23,13 +23,16 @@
 #include "absl/meta/type_traits.h"
 #include "absl/strings/match.h"
 
+#include "src/core/lib/gprpp/time.h"
 #include "src/core/lib/iomgr/error.h"
 #include "src/core/lib/slice/slice.h"
 #include "src/core/lib/surface/validate_metadata.h"
-#include "src/core/lib/transport/metadata.h"
 
 namespace grpc_core {
 
+using MetadataParseErrorFn =
+    absl::FunctionRef<void(absl::string_view error, const Slice& value)>;
+
 namespace metadata_detail {
 
 // Helper to determine whether a traits metadata is inlinable inside a memento,
@@ -37,16 +40,16 @@ namespace metadata_detail {
 template <typename Which>
 struct HasSimpleMemento {
   static constexpr bool value =
-      std::is_trivial<typename Which::MementoType>::value &&
-      sizeof(typename Which::MementoType) <= sizeof(uint64_t);
+      (std::is_trivial<typename Which::MementoType>::value &&
+       sizeof(typename Which::MementoType) <= sizeof(grpc_slice)) ||
+      std::is_same<typename Which::MementoType, Duration>::value;
 };
 
 // Storage type for a single metadata entry.
 union Buffer {
-  uint64_t trivial;
+  uint8_t trivial[sizeof(grpc_slice)];
   void* pointer;
   grpc_slice slice;
-  grpc_mdelem mdelem;
 };
 
 // Given a key and a value, concatenate together to make a debug string.
@@ -71,7 +74,9 @@ GPR_ATTRIBUTE_NOINLINE std::string MakeDebugStringPipeline(
 // Extract a trivial field value from a Buffer - for MakeDebugStringPipeline.
 template <typename Field>
 Field FieldFromTrivial(const Buffer& value) {
-  return static_cast<Field>(value.trivial);
+  Field field;
+  memcpy(&field, value.trivial, sizeof(Field));
+  return field;
 }
 
 // Extract a pointer field value from a Buffer - for MakeDebugStringPipeline.
@@ -86,6 +91,15 @@ Slice SliceFromBuffer(const Buffer& buffer);
 // Unref the grpc_slice part of a Buffer (assumes it is in fact a grpc_slice).
 void DestroySliceValue(const Buffer& value);
 
+// Destroy a trivial memento (empty function).
+void DestroyTrivialMemento(const Buffer& value);
+
+// Set a slice value in a container
+template <Slice (*MementoToValue)(Slice)>
+void SetSliceValue(Slice* set, const Buffer& value) {
+  *set = MementoToValue(SliceFromBuffer(value));
+}
+
 }  // namespace metadata_detail
 
 // A parsed metadata value.
@@ -107,7 +121,7 @@ class ParsedMetadata {
       uint32_t transport_size)
       : vtable_(ParsedMetadata::template TrivialTraitVTable<Which>()),
         transport_size_(transport_size) {
-    value_.trivial = static_cast<uint64_t>(value);
+    memcpy(value_.trivial, &value, sizeof(value));
   }
   template <typename Which>
   ParsedMetadata(
@@ -129,13 +143,12 @@ class ParsedMetadata {
         transport_size_(transport_size) {
     value_.slice = value.TakeCSlice();
   }
-  // Takes ownership of elem
-  explicit ParsedMetadata(grpc_mdelem elem)
-      : vtable_(grpc_is_binary_header_internal(GRPC_MDKEY(elem))
-                    ? MdelemVtable<true>()
-                    : MdelemVtable<false>()),
-        transport_size_(GRPC_MDELEM_LENGTH(elem)) {
-    value_.mdelem = elem;
+  // Construct metadata from a string key, slice value pair.
+  ParsedMetadata(Slice key, Slice value)
+      : vtable_(ParsedMetadata::KeyValueVTable(key.as_string_view())),
+        transport_size_(key.size() + value.size() + 32) {
+    value_.pointer =
+        new std::pair<Slice, Slice>(std::move(key), std::move(value));
   }
   ParsedMetadata() : vtable_(EmptyVTable()), transport_size_(0) {}
   ~ParsedMetadata() { vtable_->destroy(value_); }
@@ -158,9 +171,8 @@ class ParsedMetadata {
   }
 
   // Set this parsed value on a container.
-  GRPC_MUST_USE_RESULT grpc_error_handle
-  SetOnContainer(MetadataContainer* container) const {
-    return vtable_->set(value_, container);
+  void SetOnContainer(MetadataContainer* container) const {
+    vtable_->set(value_, container);
   }
 
   // Is this a binary header or not?
@@ -168,16 +180,20 @@ class ParsedMetadata {
   // HTTP2 defined storage size of this metadatum.
   uint32_t transport_size() const { return transport_size_; }
   // Create a new parsed metadata with the same key but a different value.
-  ParsedMetadata WithNewValue(Slice value) const {
+  ParsedMetadata WithNewValue(Slice value,
+                              MetadataParseErrorFn on_error) const {
     ParsedMetadata result;
     result.vtable_ = vtable_;
     result.value_ = value_;
-    result.transport_size_ =
-        TransportSize(vtable_->key_length(value_), value.length());
-    vtable_->with_new_value(&value, &result);
+    result.transport_size_ = TransportSize(key().length(), value.length());
+    vtable_->with_new_value(&value, on_error, &result);
     return result;
   }
   std::string DebugString() const { return vtable_->debug_string(value_); }
+  absl::string_view key() const {
+    if (vtable_->key == nullptr) return vtable_->key_value;
+    return vtable_->key(value_);
+  }
 
   // TODO(ctiller): move to transport
   static uint32_t TransportSize(uint32_t key_size, uint32_t value_size) {
@@ -191,29 +207,38 @@ class ParsedMetadata {
   struct VTable {
     const bool is_binary_header;
     void (*const destroy)(const Buffer& value);
-    grpc_error_handle (*const set)(const Buffer& value,
-                                   MetadataContainer* container);
+    void (*const set)(const Buffer& value, MetadataContainer* container);
     // result is a bitwise copy of the originating ParsedMetadata.
-    void (*const with_new_value)(Slice* new_value, ParsedMetadata* result);
+    void (*const with_new_value)(Slice* new_value,
+                                 MetadataParseErrorFn on_error,
+                                 ParsedMetadata* result);
     std::string (*const debug_string)(const Buffer& value);
-    // TODO(ctiller): when we delete mdelem, make this a simple integer constant
-    // on the vtable
-    size_t (*const key_length)(const Buffer& value);
+    // The key - if key is null, use key_value, otherwise call key.
+    absl::string_view key_value;
+    absl::string_view (*const key)(const Buffer& value);
   };
 
   static const VTable* EmptyVTable();
+  static const VTable* KeyValueVTable(absl::string_view key);
   template <typename Which>
   static const VTable* TrivialTraitVTable();
   template <typename Which>
   static const VTable* NonTrivialTraitVTable();
   template <typename Which>
   static const VTable* SliceTraitVTable();
-  template <bool kIsBinaryHeader>
-  static const VTable* MdelemVtable();
 
-  template <Slice (*ParseMemento)(Slice)>
-  GPR_ATTRIBUTE_NOINLINE void WithNewValueSetSlice(Slice* slice) {
-    value_.slice = ParseMemento(std::move(*slice)).TakeCSlice();
+  template <Slice (*ParseMemento)(Slice, MetadataParseErrorFn)>
+  GPR_ATTRIBUTE_NOINLINE static void WithNewValueSetSlice(
+      Slice* slice, MetadataParseErrorFn on_error, ParsedMetadata* result) {
+    result->value_.slice =
+        ParseMemento(std::move(*slice), on_error).TakeCSlice();
+  }
+
+  template <typename T, T (*ParseMemento)(Slice, MetadataParseErrorFn)>
+  GPR_ATTRIBUTE_NOINLINE static void WithNewValueSetTrivial(
+      Slice* slice, MetadataParseErrorFn on_error, ParsedMetadata* result) {
+    T memento = ParseMemento(std::move(*slice), on_error);
+    memcpy(result->value_.trivial, &memento, sizeof(memento));
   }
 
   const VTable* vtable_;
@@ -229,15 +254,16 @@ ParsedMetadata<MetadataContainer>::EmptyVTable() {
   static const VTable vtable = {
       false,
       // destroy
-      [](const Buffer&) {},
+      metadata_detail::DestroyTrivialMemento,
       // set
-      [](const Buffer&, MetadataContainer*) { return GRPC_ERROR_NONE; },
+      [](const Buffer&, MetadataContainer*) {},
       // with_new_value
-      [](Slice*, ParsedMetadata*) {},
+      [](Slice*, MetadataParseErrorFn, ParsedMetadata*) {},
       // debug_string
       [](const Buffer&) -> std::string { return "empty"; },
-      // key_length
-      [](const Buffer&) -> size_t { return 0; },
+      // key
+      "",
+      nullptr,
   };
   return &vtable;
 }
@@ -249,18 +275,17 @@ ParsedMetadata<MetadataContainer>::TrivialTraitVTable() {
   static const VTable vtable = {
       absl::EndsWith(Which::key(), "-bin"),
       // destroy
-      [](const Buffer&) {},
+      metadata_detail::DestroyTrivialMemento,
       // set
       [](const Buffer& value, MetadataContainer* map) {
-        map->Set(Which(),
-                 Which::MementoToValue(
-                     static_cast<typename Which::MementoType>(value.trivial)));
-        return GRPC_ERROR_NONE;
+        map->Set(
+            Which(),
+            Which::MementoToValue(
+                metadata_detail::FieldFromTrivial<typename Which::MementoType>(
+                    value)));
       },
       // with_new_value
-      [](Slice* value, ParsedMetadata* result) {
-        result->value_.trivial = Which::ParseMemento(std::move(*value));
-      },
+      WithNewValueSetTrivial<typename Which::MementoType, Which::ParseMemento>,
       // debug_string
       [](const Buffer& value) {
         return metadata_detail::MakeDebugStringPipeline(
@@ -268,8 +293,9 @@ ParsedMetadata<MetadataContainer>::TrivialTraitVTable() {
             metadata_detail::FieldFromTrivial<typename Which::MementoType>,
             Which::DisplayValue);
       },
-      // key_length
-      [](const Buffer&) { return Which::key().size(); },
+      // key
+      Which::key(),
+      nullptr,
   };
   return &vtable;
 }
@@ -288,12 +314,11 @@ ParsedMetadata<MetadataContainer>::NonTrivialTraitVTable() {
       [](const Buffer& value, MetadataContainer* map) {
         auto* p = static_cast<typename Which::MementoType*>(value.pointer);
         map->Set(Which(), Which::MementoToValue(*p));
-        return GRPC_ERROR_NONE;
       },
       // with_new_value
-      [](Slice* value, ParsedMetadata* result) {
-        result->value_.pointer = new
-            typename Which::MementoType(Which::ParseMemento(std::move(*value)));
+      [](Slice* value, MetadataParseErrorFn on_error, ParsedMetadata* result) {
+        result->value_.pointer = new typename Which::MementoType(
+            Which::ParseMemento(std::move(*value), on_error));
       },
       // debug_string
       [](const Buffer& value) {
@@ -302,8 +327,9 @@ ParsedMetadata<MetadataContainer>::NonTrivialTraitVTable() {
             metadata_detail::FieldFromPointer<typename Which::MementoType>,
             Which::DisplayValue);
       },
-      // key_length
-      [](const Buffer&) { return Which::key().size(); },
+      // key
+      Which::key(),
+      nullptr,
   };
   return &vtable;
 }
@@ -318,63 +344,56 @@ ParsedMetadata<MetadataContainer>::SliceTraitVTable() {
       metadata_detail::DestroySliceValue,
       // set
       [](const Buffer& value, MetadataContainer* map) {
-        map->Set(Which(), Which::MementoToValue(
-                              metadata_detail::SliceFromBuffer(value)));
-        return GRPC_ERROR_NONE;
+        metadata_detail::SetSliceValue<Which::MementoToValue>(
+            map->GetOrCreatePointer(Which()), value);
       },
       // with_new_value
-      [](Slice* value, ParsedMetadata* result) {
-        result->WithNewValueSetSlice<Which::ParseMemento>(value);
-      },
+      WithNewValueSetSlice<Which::ParseMemento>,
       // debug_string
       [](const Buffer& value) {
         return metadata_detail::MakeDebugStringPipeline(
             Which::key(), value, metadata_detail::SliceFromBuffer,
             Which::DisplayValue);
       },
-      // key_length
-      [](const Buffer&) { return Which::key().size(); },
+      // key
+      Which::key(),
+      nullptr,
   };
   return &vtable;
 }
 
 template <typename MetadataContainer>
-template <bool kIsBinaryHeader>
 const typename ParsedMetadata<MetadataContainer>::VTable*
-ParsedMetadata<MetadataContainer>::MdelemVtable() {
-  static const VTable vtable = {
-      kIsBinaryHeader,
-      // destroy
-      [](const Buffer& value) { GRPC_MDELEM_UNREF(value.mdelem); },
-      // set
-      [](const Buffer& value, MetadataContainer* map) {
-        auto md = GRPC_MDELEM_REF(value.mdelem);
-        auto err = map->Append(md);
-        // If an error occurs, md is not consumed and we need to.
-        // This is an awful API, but that's why we're replacing it.
-        if (err != GRPC_ERROR_NONE) {
-          GRPC_MDELEM_UNREF(md);
-        }
-        return err;
-      },
-      // with_new_value
-      [](Slice* value_slice, ParsedMetadata* result) {
-        result->value_.mdelem = grpc_mdelem_from_slices(
-            static_cast<const ManagedMemorySlice&>(
-                grpc_slice_ref_internal(GRPC_MDKEY(result->value_.mdelem))),
-            value_slice->TakeCSlice());
-      },
-      // debug_string
-      [](const Buffer& value) {
-        return metadata_detail::MakeDebugString(
-            StringViewFromSlice(GRPC_MDKEY(value.mdelem)),
-            StringViewFromSlice(GRPC_MDVALUE(value.mdelem)));
-      },
-      // key_length
-      [](const Buffer& value) {
-        return GRPC_SLICE_LENGTH(GRPC_MDKEY(value.mdelem));
-      }};
-  return &vtable;
+ParsedMetadata<MetadataContainer>::KeyValueVTable(absl::string_view key) {
+  using KV = std::pair<Slice, Slice>;
+  static const auto destroy = [](const Buffer& value) {
+    delete static_cast<KV*>(value.pointer);
+  };
+  static const auto set = [](const Buffer& value, MetadataContainer* map) {
+    auto* p = static_cast<KV*>(value.pointer);
+    map->AppendUnknown(p->first.as_string_view(), p->second.Ref());
+  };
+  static const auto with_new_value = [](Slice* value, MetadataParseErrorFn,
+                                        ParsedMetadata* result) {
+    auto* p = new KV{
+        static_cast<KV*>(result->value_.pointer)->first.Ref(),
+        std::move(*value),
+    };
+    result->value_.pointer = p;
+  };
+  static const auto debug_string = [](const Buffer& value) {
+    auto* p = static_cast<KV*>(value.pointer);
+    return absl::StrCat(p->first.as_string_view(), ": ",
+                        p->second.as_string_view());
+  };
+  static const auto key_fn = [](const Buffer& value) {
+    return static_cast<KV*>(value.pointer)->first.as_string_view();
+  };
+  static const VTable vtable[2] = {
+      {false, destroy, set, with_new_value, debug_string, "", key_fn},
+      {true, destroy, set, with_new_value, debug_string, "", key_fn},
+  };
+  return &vtable[absl::EndsWith(key, "-bin")];
 }
 
 }  // namespace grpc_core
